@@ -1,9 +1,9 @@
-/*  Lziprecover - Data recovery tool for lzip files
-    Copyright (C) 2009, 2010, 2011, 2012, 2013 Antonio Diaz Diaz.
+/*  Lziprecover - Data recovery tool for the lzip format
+    Copyright (C) 2009-2017 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
+    the Free Software Foundation, either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
@@ -47,28 +47,31 @@ public:
 
   ~Range_decoder() { delete[] buffer; }
 
-  bool code_is_zero() const { return ( code == 0 ); }
+  unsigned get_code() const { return code; }
   bool finished() { return pos >= stream_pos && !read_block(); }
   unsigned long long member_position() const { return partial_member_pos + pos; }
-  void reset_member_position() { partial_member_pos = -pos; }
+
+  void reset_member_position()
+    { partial_member_pos = 0; partial_member_pos -= pos; }
 
   uint8_t get_byte()
     {
-    if( finished() ) return 0xAA;		// make code != 0
+    // 0xFF avoids decoder error if member is truncated at EOS marker
+    if( finished() ) return 0xFF;
     return buffer[pos++];
     }
 
   int read_data( uint8_t * const outbuf, const int size )
     {
-    int rest = size;
-    while( rest > 0 && !finished() )
+    int sz = 0;
+    while( sz < size && !finished() )
       {
-      const int rd = std::min( rest, stream_pos - pos );
-      std::memcpy( outbuf + size - rest, buffer + pos, rd );
+      const int rd = std::min( size - sz, stream_pos - pos );
+      std::memcpy( outbuf + sz, buffer + pos, rd );
       pos += rd;
-      rest -= rd;
+      sz += rd;
       }
-    return size - rest;
+    return sz;
     }
 
   void load()
@@ -76,6 +79,7 @@ public:
     code = 0;
     for( int i = 0; i < 5; ++i ) code = (code << 8) | get_byte();
     range = 0xFFFFFFFFU;
+    code &= range;		// make sure that first byte is discarded
     }
 
   void normalize()
@@ -84,24 +88,23 @@ public:
       { range <<= 8; code = (code << 8) | get_byte(); }
     }
 
-  int decode( const int num_bits )
+  unsigned decode( const int num_bits )
     {
-    int symbol = 0;
+    unsigned symbol = 0;
     for( int i = num_bits; i > 0; --i )
       {
       normalize();
       range >>= 1;
 //      symbol <<= 1;
 //      if( code >= range ) { code -= range; symbol |= 1; }
-      const uint32_t mask = 0U - (code < range);
-      code -= range;
-      code += range & mask;
-      symbol = (symbol << 1) + (mask + 1);
+      const bool bit = ( code >= range );
+      symbol = ( symbol << 1 ) + bit;
+      code -= range & ( 0U - bit );
       }
     return symbol;
     }
 
-  int decode_bit( Bit_model & bm )
+  unsigned decode_bit( Bit_model & bm )
     {
     normalize();
     const uint32_t bound = ( range >> bit_model_total_bits ) * bm.probability;
@@ -120,62 +123,68 @@ public:
       }
     }
 
-  int decode_tree( Bit_model bm[], const int num_bits )
+  unsigned decode_tree3( Bit_model bm[] )
     {
-    int symbol = 1;
-    for( int i = num_bits; i > 0; --i )
+    unsigned symbol = 1;
+    symbol = ( symbol << 1 ) | decode_bit( bm[symbol] );
+    symbol = ( symbol << 1 ) | decode_bit( bm[symbol] );
+    symbol = ( symbol << 1 ) | decode_bit( bm[symbol] );
+    return symbol & 7;
+    }
+
+  unsigned decode_tree6( Bit_model bm[] )
+    {
+    unsigned symbol = 1;
+    symbol = ( symbol << 1 ) | decode_bit( bm[symbol] );
+    symbol = ( symbol << 1 ) | decode_bit( bm[symbol] );
+    symbol = ( symbol << 1 ) | decode_bit( bm[symbol] );
+    symbol = ( symbol << 1 ) | decode_bit( bm[symbol] );
+    symbol = ( symbol << 1 ) | decode_bit( bm[symbol] );
+    symbol = ( symbol << 1 ) | decode_bit( bm[symbol] );
+    return symbol & 0x3F;
+    }
+
+  unsigned decode_tree8( Bit_model bm[] )
+    {
+    unsigned symbol = 1;
+    for( int i = 0; i < 8; ++i )
       symbol = ( symbol << 1 ) | decode_bit( bm[symbol] );
-    return symbol - (1 << num_bits);
+    return symbol & 0xFF;
     }
 
-  int decode_tree6( Bit_model bm[] )
+  unsigned decode_tree_reversed( Bit_model bm[], const int num_bits )
     {
-    int symbol = 1;
-    symbol = ( symbol << 1 ) | decode_bit( bm[symbol] );
-    symbol = ( symbol << 1 ) | decode_bit( bm[symbol] );
-    symbol = ( symbol << 1 ) | decode_bit( bm[symbol] );
-    symbol = ( symbol << 1 ) | decode_bit( bm[symbol] );
-    symbol = ( symbol << 1 ) | decode_bit( bm[symbol] );
-    symbol = ( symbol << 1 ) | decode_bit( bm[symbol] );
-    return symbol - (1 << 6);
-    }
-
-  int decode_tree_reversed( Bit_model bm[], const int num_bits )
-    {
-    int model = 1;
-    int symbol = 0;
+    unsigned model = 1;
+    unsigned symbol = 0;
     for( int i = 0; i < num_bits; ++i )
       {
-      const bool bit = decode_bit( bm[model] );
-      model <<= 1;
-      if( bit ) { ++model; symbol |= (1 << i); }
+      const unsigned bit = decode_bit( bm[model] );
+      model = ( model << 1 ) + bit;
+      symbol |= ( bit << i );
       }
     return symbol;
     }
 
-  int decode_tree_reversed4( Bit_model bm[] )
+  unsigned decode_tree_reversed4( Bit_model bm[] )
     {
-    int model = 1;
-    int symbol = 0;
-    int bit = decode_bit( bm[model] );
-    model = (model << 1) + bit; symbol |= bit;
+    unsigned symbol = decode_bit( bm[1] );
+    unsigned model = 2 + symbol;
+    unsigned bit = decode_bit( bm[model] );
+    model = ( model << 1 ) + bit; symbol |= ( bit << 1 );
     bit = decode_bit( bm[model] );
-    model = (model << 1) + bit; symbol |= (bit << 1);
-    bit = decode_bit( bm[model] );
-    model = (model << 1) + bit; symbol |= (bit << 2);
-    if( decode_bit( bm[model] ) ) symbol |= 8;
+    model = ( model << 1 ) + bit; symbol |= ( bit << 2 );
+    symbol |= ( decode_bit( bm[model] ) << 3 );
     return symbol;
     }
 
-  int decode_matched( Bit_model bm[], int match_byte )
+  unsigned decode_matched( Bit_model bm[], unsigned match_byte )
     {
     Bit_model * const bm1 = bm + 0x100;
-    int symbol = 1;
-    for( int i = 7; i >= 0; --i )
+    unsigned symbol = 1;
+    while( symbol < 0x100 )
       {
-      match_byte <<= 1;
-      const int match_bit = match_byte & 0x100;
-      const int bit = decode_bit( bm1[match_bit+symbol] );
+      const unsigned match_bit = ( match_byte <<= 1 ) & 0x100;
+      const unsigned bit = decode_bit( bm1[match_bit+symbol] );
       symbol = ( symbol << 1 ) | bit;
       if( match_bit != bit << 8 )
         {
@@ -184,18 +193,16 @@ public:
         break;
         }
       }
-    return symbol - 0x100;
+    return symbol & 0xFF;
     }
 
-  int decode_len( Len_model & lm, const int pos_state )
+  unsigned decode_len( Len_model & lm, const int pos_state )
     {
     if( decode_bit( lm.choice1 ) == 0 )
-      return decode_tree( lm.bm_low[pos_state], len_low_bits );
+      return decode_tree3( lm.bm_low[pos_state] );
     if( decode_bit( lm.choice2 ) == 0 )
-      return len_low_symbols +
-             decode_tree( lm.bm_mid[pos_state], len_mid_bits );
-    return len_low_symbols + len_mid_symbols +
-           decode_tree( lm.bm_high, len_high_bits );
+      return len_low_symbols + decode_tree3( lm.bm_mid[pos_state] );
+    return len_low_symbols + len_mid_symbols + decode_tree8( lm.bm_high );
     }
   };
 
@@ -206,52 +213,67 @@ class LZ_decoder
   const unsigned long long outend;
   unsigned long long partial_data_pos;
   Range_decoder & rdec;
-  const int dictionary_size;
-  const int buffer_size;
+  const unsigned dictionary_size;
   uint8_t * const buffer;	// output buffer
-  int pos;			// current pos in buffer
-  int stream_pos;		// first byte not yet written to file
+  unsigned pos;			// current pos in buffer
+  unsigned stream_pos;		// first byte not yet written to file
   uint32_t crc_;
   const int outfd;		// output file descriptor
+  bool pos_wrapped;
 
   unsigned long long stream_position() const
     { return partial_data_pos + stream_pos; }
   void flush_data();
   bool verify_trailer( const Pretty_print & pp ) const;
 
-  uint8_t get_prev_byte() const
+  uint8_t peek_prev() const
     {
-    const int i = ( ( pos > 0 ) ? pos : buffer_size ) - 1;
-    return buffer[i];
+    if( pos > 0 ) return buffer[pos-1];
+    if( pos_wrapped ) return buffer[dictionary_size-1];
+    return 0;			// prev_byte of first byte
     }
 
-  uint8_t get_byte( const int distance ) const
+  uint8_t peek( const unsigned distance ) const
     {
-    int i = pos - distance - 1;
-    if( i < 0 ) i += buffer_size;
+    const unsigned i = ( ( pos > distance ) ? 0 : dictionary_size ) +
+                       pos - distance - 1;
     return buffer[i];
     }
 
   void put_byte( const uint8_t b )
     {
     buffer[pos] = b;
-    if( ++pos >= buffer_size ) flush_data();
+    if( ++pos >= dictionary_size ) flush_data();
     }
 
-  void copy_block( const int distance, int len )
+  void copy_block( const unsigned distance, unsigned len )
     {
-    int i = pos - distance - 1;
-    if( i < 0 ) i += buffer_size;
-    if( len < buffer_size - std::max( pos, i ) && len <= std::abs( pos - i ) )
+    unsigned lpos = pos, i = lpos - distance - 1;
+    bool fast, fast2;
+    if( lpos > distance )
+      {
+      fast = ( len < dictionary_size - lpos );
+      fast2 = ( fast && len <= lpos - i );
+      }
+    else
+      {
+      i += dictionary_size;
+      fast = ( len < dictionary_size - i );	// (i == pos) may happen
+      fast2 = ( fast && len <= i - lpos );
+      }
+    if( fast )					// no wrap
       {
-      std::memcpy( buffer + pos, buffer + i, len );	// no wrap, no overlap
       pos += len;
+      if( fast2 )				// no wrap, no overlap
+        std::memcpy( buffer + lpos, buffer + i, len );
+      else
+        for( ; len > 0; --len ) buffer[lpos++] = buffer[i++];
       }
     else for( ; len > 0; --len )
       {
       buffer[pos] = buffer[i];
-      if( ++pos >= buffer_size ) flush_data();
-      if( ++i >= buffer_size ) i = 0;
+      if( ++pos >= dictionary_size ) flush_data();
+      if( ++i >= dictionary_size ) i = 0;
       }
     }
 
@@ -259,7 +281,7 @@ class LZ_decoder
   void operator=( const LZ_decoder & );		// declared as private
 
 public:
-  LZ_decoder( const File_header & header, Range_decoder & rde, const int ofd,
+  LZ_decoder( Range_decoder & rde, const unsigned dict_size, const int ofd,
               const unsigned long long oskip = 0,
               const unsigned long long oend = -1ULL )
     :
@@ -267,14 +289,14 @@ public:
     outend( oend ),
     partial_data_pos( 0 ),
     rdec( rde ),
-    dictionary_size( header.dictionary_size() ),
-    buffer_size( std::max( 65536, dictionary_size ) ),
-    buffer( new uint8_t[buffer_size] ),
+    dictionary_size( dict_size ),
+    buffer( new uint8_t[dictionary_size] ),
     pos( 0 ),
     stream_pos( 0 ),
     crc_( 0xFFFFFFFFU ),
-    outfd( ofd )
-    { buffer[buffer_size-1] = 0; }	// prev_byte of first_byte
+    outfd( ofd ),
+    pos_wrapped( false )
+    {}
 
   ~LZ_decoder() { delete[] buffer; }
 
